# Copyright 2021 Ruffin White
# Licensed under the Apache License, Version 2.0

import os

from colcon_clean.base_handler import BaseHandlerExtensionPoint
from colcon_core.plugin_system import satisfies_version

BASE_PATH = 'install'


class InstallBaseHandler(BaseHandlerExtensionPoint):
    """Determin how install paths for the workspace should be cleaned."""

    def __init__(self):  # noqa: D107
        super().__init__(BASE_PATH)
        satisfies_version(
            BaseHandlerExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        parser.add_argument(
            '--install-base',
            default=self.base_path,
            help='The base path for all install directories '
                 f'(default: {self.base_path})')

    def get_workspace_paths(self, *, args):  # noqa: D102
        return [args.install_base]

    def get_package_paths(self, *, args, pkg):  # noqa: D102
        return [os.path.join(args.install_base, pkg.name)]



# Copyright 2016-2018 Dirk Thomas
# Copyright 2021 Ruffin White
# Licensed under the Apache License, Version 2.0

from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_by_name


class BaseHandlerExtensionPoint:
    """
    The interface for base handler extensions.

    A base handler extension determines how a base
    should cleaned for subset of packages to be processed.

    For each instance the attribute `BASE_HANDLER_NAME` is being set to
    the basename of the entry point registering the extension.
    """

    """The version of the package selection extension interface."""
    EXTENSION_POINT_VERSION = '1.0'

    def __init__(self, base_path):  # noqa: D107
        self.base_path = base_path

    def add_arguments(self, *, parser):
        """
        Add command line arguments specific to the workspace base.

        This method must be overridden in a subclass.

        :param parser: The argument parser
        """
        raise NotImplementedError()

    def get_workspace_paths(self, *, args):
        """
        Get paths for a workspace.

        This method must be overridden in a subclass.

        :param args: The parsed command line arguments

        :rtype: list
        """
        raise NotImplementedError()

    def get_package_paths(self, *, args, pkg):
        """
        Get paths for a package.

        This method must be overridden in a subclass.

        :param args: The parsed command line arguments
        :param pkg: The package descriptor

        :rtype: list
        """
        raise NotImplementedError()


def add_base_handler_arguments(parser):
    """
    Add the command line arguments for the base handler extensions.

    :param parser: The argument parser
    """
    group = parser.add_argument_group(title='Base handler arguments')
    extensions = get_base_handler_extensions()
    extension_keys = sorted(extensions.keys())

    group.add_argument(
        '--base-select', nargs='*', metavar='BASE_NAME',
        choices=extension_keys,
        default=extension_keys,
        help='Select base names to clean in workspace '
             f'(default: {extension_keys})')

    group.add_argument(
        '--base-ignore', nargs='*', metavar='BASE_NAME',
        choices=extension_keys,
        default=[],
        help='Ignore base names to clean in workspace '
             '(default: [])')

    for key in sorted(extensions.keys()):
        extension = extensions[key]
        extension.add_arguments(parser=group)


def get_base_handler_extensions():
    """
    Get the available base handler extensions.

    The extensions are ordered by their entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name, extension in extensions.items():
        extension.BASE_HANDLER_NAME = name
    return order_extensions_by_name(extensions)



# Copyright 2021 Ruffin White
# Licensed under the Apache License, Version 2.0

from unittest.mock import Mock

from colcon_clean.verb.clean import CleanVerb
from colcon_core.command import CommandContext


class Object(object):
    pass


def test_verb_interface():
    interface = CleanVerb()
    interface._subparser = Object()
    interface._subparser.format_usage = Mock(return_value='')

    args = Object()
    context = CommandContext(command_name='clean', args=args)

    context.args.subverb_name = None
    rc = interface.main(context=context)
    assert rc == 'Error: No subverb provided'

    context.args.subverb_name = 'packages'
    rc = interface.main(context=context)
    assert rc is None

    context.args.subverb_name = 'workspace'
    rc = interface.main(context=context)
    assert rc is None



# Copyright 2021 Ruffin White
# Licensed under the Apache License, Version 2.0

import argparse
import os
from pathlib import Path
import shutil
import sys
from tempfile import mkdtemp

from colcon_core.command import main
import pytest


def _raising_error(self, message):
    raise sys.exc_info()[1]


argparse.ArgumentParser.error = _raising_error


def test_main(monkeypatch):
    """System test for colcon clean CLI."""
    ws_base = Path(mkdtemp(prefix='test_colcon_'))
    resources_base = Path('test', 'resources').absolute()
    shutil.copytree(resources_base / 'test_src', ws_base / 'src')

    os.chdir(ws_base)
    argv = []
    os.environ['COLCON_EXTENSION_BLOCKLIST'] = (
        'colcon_core.event_handler.desktop_notification:' +
        os.environ.get('COLCON_EXTENSION_BLOCKLIST', ''))
    os.environ['COLCON_DEFAULTS_FILE'] = \
        (resources_base / 'defaults.yaml').as_posix()

    try:
        main(argv=argv + ['build'])
        main(argv=argv + ['test'])

        # Clean all package base paths explicitly
        main(argv=argv + ['clean', 'packages', '--yes', \
            '--base-select', \
                'build', \
                'install', \
                'log', \
                'test_result', \
            '--packages-select', \
                'test-package-b', \
                'test-package-c'])  # noqa

        # Assert unselected packages are skipped
        assert (ws_base / 'build' / 'test-package-a').exists()
        assert (ws_base / 'install' / 'test-package-a').exists()
        assert (ws_base / 'test_results' / 'test-package-a').exists()

        # Assert selected packages are cleaned
        assert not (ws_base / 'build' / 'test-package-b').exists()
        assert not (ws_base / 'build' / 'test-package-c').exists()
        assert not (ws_base / 'install' / 'test-package-b').exists()
        assert not (ws_base / 'install' / 'test-package-c').exists()
        assert not (ws_base / 'test_results' / 'test-package-b').exists()
        assert not (ws_base / 'test_results' / 'test-package-c').exists()

        # Assert inapplicable base paths are skipped
        assert (ws_base / 'log').exists()

        # Clean workspace build base paths of python files
        # And with duplicate match filters
        main(argv=argv + ['clean', 'workspace', '--yes', \
            '--base-select', \
                'build', \
            '--clean-match', \
                '*.py', \
                '*.py'])  # noqa

        # Assert workspace matches are cleaned
        assert not (ws_base / 'build' / 'test-package-a' / 'build' / 'lib' /
                    'test_package_a ' / '__init__.py').exists()

        # Assert workspace misses are not cleaned
        assert (ws_base / 'build' / 'test-package-a' /
                'colcon_build.rc').exists()

        # Try again but with nothing left to clean
        main(argv=argv + ['clean', 'workspace', '--yes', \
            '--base-select', \
                'build', \
            '--clean-match', \
                '*.py', \
                '*.py'])  # noqa

        # Clean all workspace base paths explicitly
        main(argv=argv + ['clean', 'workspace', '--yes', \
            '--base-select', \
                'build', \
                'install', \
                'log', \
                'test_result'])  # noqa

        # Assert workspace base paths are cleaned
        assert not (ws_base / 'build').exists()
        assert not (ws_base / 'install').exists()
        assert not (ws_base / 'test_results').exists()
        # Exception for log, as clean verb generates its own
        # assert not (ws_base / 'log').exists()

        # Try again implicitly but with nothing left to clean
        main(argv=argv + ['clean', 'workspace', '--yes'])  # noqa

        main(argv=argv + ['build'])
        main(argv=argv + ['test'])

        # Don't clean workspace base paths when prompted by user input
        monkeypatch.setattr('builtins.input', lambda: 'n')
        main(argv=argv + ['clean', 'workspace'])  # noqa

        # Ignore one workspace base paths explicitly
        main(argv=argv + ['clean', 'workspace', \
            '--base-ignore', \
                'log'])  # noqa

        # Assert no workspace base paths are cleaned
        assert (ws_base / 'build').exists()
        assert (ws_base / 'install').exists()
        assert (ws_base / 'test_results').exists()
        assert (ws_base / 'log').exists()

        # Ignore one package base paths explicitly
        main(argv=argv + ['clean', 'packages', \
            '--base-ignore', \
                'log'])  # noqa

        # Assert no package base paths are cleaned
        assert (ws_base / 'build' / 'test-package-a').exists()
        assert (ws_base / 'build' / 'test-package-b').exists()
        assert (ws_base / 'build' / 'test-package-c').exists()
        assert (ws_base / 'install' / 'test-package-a').exists()
        assert (ws_base / 'install' / 'test-package-b').exists()
        assert (ws_base / 'install' / 'test-package-c').exists()
        assert (ws_base / 'test_results' / 'test-package-a').exists()
        assert (ws_base / 'test_results' / 'test-package-b').exists()
        assert (ws_base / 'test_results' / 'test-package-c').exists()
        assert (ws_base / 'log').exists()

        # Clean workspace base paths when prompted by user input
        monkeypatch.setattr('builtins.input', lambda: 'y')

        # Ignore one package explicitly
        main(argv=argv + ['clean', 'packages', \
            '--packages-skip', \
                'test-package-a'])  # noqa

        # Assert unselected packages are skipped
        assert (ws_base / 'build' / 'test-package-a').exists()
        assert (ws_base / 'install' / 'test-package-a').exists()
        assert (ws_base / 'test_results' / 'test-package-a').exists()

        # Assert selected packages are cleaned
        assert not (ws_base / 'build' / 'test-package-b').exists()
        assert not (ws_base / 'build' / 'test-package-c').exists()
        assert not (ws_base / 'install' / 'test-package-b').exists()
        assert not (ws_base / 'install' / 'test-package-c').exists()
        assert not (ws_base / 'test_results' / 'test-package-b').exists()
        assert not (ws_base / 'test_results' / 'test-package-c').exists()

        # Ignore one workspace base paths explicitly
        main(argv=argv + ['clean', 'workspace', \
            '--base-ignore', \
                'log'])  # noqa

        # Assert implicitly selected workspace base paths are cleaned
        assert not (ws_base / 'build').exists()
        assert not (ws_base / 'install').exists()
        assert not (ws_base / 'test_results').exists()
        assert (ws_base / 'log').exists()

        main(argv=argv + ['clean', 'workspace'])  # noqa

        with pytest.raises(argparse.ArgumentError):
            # Try cleaning packages with invalid base handler selection
            main(argv=argv + ['clean', 'packages', \
                '--base-select', \
                    'foo'])  # noqa

        with pytest.raises(argparse.ArgumentError):
            # Try cleaning workspace with invalid base handler selection
            main(argv=argv + ['clean', 'workspace', \
                '--base-select', \
                    'bar'])  # noqa

        print('ws_base: ', ws_base)
    finally:
        # the logging subsystem might still have file handles pending
        # therefore only try to delete the temporary directory
        # shutil.rmtree(ws_base, ignore_errors=True)
        pass



# Copyright 2016-2018 Dirk Thomas
# Copyright 2021 Ruffin White
# Licensed under the Apache License, Version 2.0

import os
from pathlib import Path
import shutil

from colcon_clean.clean.query import query_yes_no
from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_by_name
from scantree import RecursionFilter, scantree


logger = colcon_logger.getChild(__name__)


class CleanSubverbExtensionPoint:
    """
    The interface for clean subverb extensions.

    A clean subverb extension provides a subverb to the `clean` verb of
    the command line tool.

    For each instance the attribute `SUBVERB_NAME` is being set to the basename
    of the entry point registering the extension.
    """

    """The version of the clean subverb extension interface."""
    EXTENSION_POINT_VERSION = '1.0'

    def add_arguments(self, *, parser):
        """
        Add command line arguments specific to the clean subverb.

        The method is intended to be overridden in a subclass.

        :param parser: The argument parser
        """
        pass

    def main(self, *, context):
        """
        Execute the clean subverb extension logic.

        This method must be overridden in a subclass.

        :param context: The context providing the parsed command line arguments
        :returns: The return code
        """
        raise NotImplementedError()


def add_clean_subverb_arguments(parser):
    """
    Add the command line arguments for the clean subverb extensions.

    :param parser: The argument parser
    """
    group = parser.add_argument_group(title='Clean subverb arguments')

    group.add_argument(
        '-y', '--yes',
        action='store_true',
        help='Automatic yes to prompts')

    filter_options = parser.add_argument_group(
        title='Clean filter arguments',
        description='Specify what files and directories to include. All '
        'files and directories (including symbolic links) are included '
        'by default. The --clean-match/--clean-ignore arguments '
        'allows for selection using glob/wildcard (".gitignore style") '
        'path matching. Paths relative to the root `directory` (i.e. '
        'excluding the name of the root directory itself) are matched '
        'against the provided patterns. For example, to only include '
        'Gcov Data files, use: `colcon clean workspace --clean-match '
        '"*.gcda"` or to exclude hidden files and directories use: '
        '`colcon clean workspace --clean-ignore ".*" ".*/"` '
        'which is short for '
        '`colcon clean workspace --clean-match "*" "!.*" "!.*/"`. '
    )
    # # FIXME: Find out how colcon help could display group description
    # filter_options = parser

    filter_options.add_argument(
        '--clean-match',
        nargs='+',
        default=None,
        help='One or several patterns for paths to include. NOTE: '
        'patterns with an asterisk must be in quotes ("*") or the '
        'asterisk preceded by an escape character (\\*).',
        metavar=''
    )
    filter_options.add_argument(
        '--clean-ignore',
        nargs='+',
        default=None,
        help='One or several patterns for paths to exclude. NOTE: '
        'patterns with an asterisk must be in quotes ("*") or the '
        'asterisk preceded by an escape character (\\*).',
        metavar=''
    )
    filter_options.add_argument(
        '--clean-no-linked-dirs',
        action='store_false',
        help='Do not include symbolic links to other directories.'
    )
    filter_options.add_argument(
        '--clean-no-linked-files',
        action='store_false',
        help='Do not include symbolic links to files.'
    )
    filter_options.set_defaults(
        clean_no_linked_dirs=True,
        clean_no_linked_files=True,
    )


def scan_directory(directory, recursion_filter):
    """
    Scan directory with recursion filter.

    The recursion filter includes match patterns or is None.

    :param directory: Path
    :param recursion_filter: RecursionFilter

    :rtype: list
    """
    base_paths = set()
    if not directory.exists():
        return base_paths

    if recursion_filter:
        tree = scantree(
            directory=directory,
            recursion_filter=recursion_filter,
            follow_links=False,
            include_empty=True)
        for filepath in tree.filepaths():
            filepath = Path(filepath)
            if directory in filepath.parents:  # pragma: no branch
                base_paths.add(filepath)
    else:
        base_paths.add(directory)
    return base_paths


def get_recursion_filter(args):
    """
    Get the recursion filter.

    The recursion filter includes match patterns or is None.

    :param args: The parsed command line arguments

    :rtype: RecursionFilter or None
    """
    if args.clean_match or args.clean_ignore:
        match_patterns = get_match_patterns(
            match=args.clean_match,
            ignore=args.clean_ignore)
        recursion_filter = RecursionFilter(
            linked_dirs=args.clean_no_linked_dirs,
            linked_files=args.clean_no_linked_files,
            match=match_patterns
        )
        return recursion_filter
    return None


def get_match_patterns(
    match=None,
    ignore=None
):
    """
    Compose match pattern from list of glob/wildcard (".gitignore style").

    # Arguments
        match: Optional[List[str]] - A list of match-patterns for files to
            *include*. Default `None` which is equivalent to `['*']`,
            i.e. everything is included (unless excluded by arguments below).
        ignore: Optional[List[str]] -  A list of match-patterns for files to
            *ignore*. Default `None` (no ignore patterns).
    """
    match = ['*'] if match is None else list(match)
    ignore = [] if ignore is None else list(ignore)

    match_spec = match + ['!' + ign for ign in ignore]

    def deduplicate(items):
        items_set = set()
        dd_items = []
        for item in items:
            if item not in items_set:
                dd_items.append(item)
                items_set.add(item)

        return dd_items

    return deduplicate(match_spec)


def get_subverb_extensions():
    """
    Get the available subverb extensions.

    The extensions are ordered by their entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name, extension in extensions.items():
        extension.SUBVERB_NAME = name
    return order_extensions_by_name(extensions)


def clean_paths(paths, confirmed=False):
    """
    Clean provided paths with conformation.

    :paths: list
    :confirmed: bool
    """
    if not paths:
        message = 'No paths cleaned.'
        logger.info(message)
        print(message)
        return

    cwd_path = Path.cwd()
    if not confirmed:
        print('Paths:')
        relpaths = (os.path.relpath(path, cwd_path) for path in paths)
        for path in sorted(relpaths):
            print('    ', path)
        question = 'Clean the above paths?'
        confirmed = query_yes_no(question)

    if confirmed:
        for path in sorted(paths):
            _clean_path(path)


def _onerror(func, path, excinfo):  # pragma: no cover
    if excinfo[0] in (OSError, PermissionError):  # pragma: no branch
        logger.warning(f"Skipping path: '{path}'")
        logger.info(f"Skipping info: '{excinfo[1]}'")
        return
    raise


def _onexc(func, path, excinfo):  # pragma: no cover
    if isinstance(excinfo, (PermissionError, OSError)):  # pragma: no branch
        logger.warning(f"Skipping path: '{path}'")
        logger.info(f"Skipping info: '{excinfo}'")
        return
    raise


def _clean_path(path):
    logger.info(f"Cleaning path: '{path}'")
    if path.is_dir():
        try:
            shutil.rmtree(path, onexc=_onexc)
        except TypeError:
            # TODO: Remove when minimum python version is 3.12
            shutil.rmtree(path, onerror=_onerror)
    else:
        path.unlink()



# Copyright 2021 Ruffin White
# Licensed under the Apache License, Version 2.0

from colcon_clean.clean.query import query_yes_no
import pytest


def test_query_interface(monkeypatch):
    question = 'Clean the above paths?'

    monkeypatch.setattr('builtins.input', lambda: '')
    assert query_yes_no(question, default='yes') is True
    assert query_yes_no(question, default='no') is False
    with pytest.raises(ValueError):
        query_yes_no(question, default='foo')

    input_iter = iter(['', 'y'])
    monkeypatch.setattr('builtins.input', lambda x=input_iter: next(x))
    assert query_yes_no(question, default=None) is True

    input_iter = iter(['', 'n'])
    monkeypatch.setattr('builtins.input', lambda x=input_iter: next(x))
    assert query_yes_no(question, default=None) is False



# Copyright 2021 Ruffin White
# Licensed under the Apache License, Version 2.0

import sys


def query_yes_no(question, default='yes'):
    """Ask a yes/no question via raw_input() and return their answer.

    https://stackoverflow.com/a/3041990/2577586
    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
            It must be "yes" (the default), "no" or None (meaning
            an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {'yes': True, 'y': True, 'ye': True, 'no': False, 'n': False}
    if default is None:
        prompt = ' [y/n] '
    elif default == 'yes':
        prompt = ' [Y/n] '
    elif default == 'no':
        prompt = ' [y/N] '
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write(
                "Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n")



# Copyright 2016-2018 Dirk Thomas
# Copyright 2021 Ruffin White
# Licensed under the Apache License, Version 2.0

from pathlib import Path

from colcon_clean.base_handler \
    import add_base_handler_arguments, get_base_handler_extensions
from colcon_clean.subverb import (
    add_clean_subverb_arguments,
    clean_paths,
    CleanSubverbExtensionPoint,
    get_recursion_filter,
    scan_directory,
)
from colcon_core.event_handler import add_event_handler_arguments
from colcon_core.plugin_system import satisfies_version
from colcon_core.verb import check_and_mark_build_tool
from colcon_core.verb import logger


class WorkspaceCleanSubverb(CleanSubverbExtensionPoint):
    """Clean current workspace."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            CleanSubverbExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        add_clean_subverb_arguments(parser)
        add_base_handler_arguments(parser)
        add_event_handler_arguments(parser)

    def main(self, *, context):  # noqa: D102
        check_and_mark_build_tool(context.args.build_base)

        base_handler_extensions = get_base_handler_extensions()
        base_paths = set()

        args = context.args
        recursion_filter = get_recursion_filter(args)

        for base_name in args.base_select:
            if base_name in args.base_ignore:
                logger.info(
                    f"Ignoring base handler for selection '{base_name}'")
                continue
            base_handler_extension = base_handler_extensions[base_name]
            workspace_paths = \
                base_handler_extension.get_workspace_paths(args=args)
            for workspace_path in workspace_paths:
                workspace_path = Path(workspace_path).absolute()
                base_paths.update(
                    scan_directory(workspace_path, recursion_filter))

        clean_paths(
            paths=base_paths,
            confirmed=args.yes)

        return 0



# Copyright 2016-2018 Dirk Thomas
# Copyright 2021 Ruffin White
# Licensed under the Apache License, Version 2.0

from colcon_clean.subverb import get_subverb_extensions
from colcon_core.command import add_subparsers
from colcon_core.plugin_system import satisfies_version
from colcon_core.verb import VerbExtensionPoint


class CleanVerb(VerbExtensionPoint):
    """Cleans package workspaces."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(VerbExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')
        self._subparser = None

    def add_arguments(self, *, parser):  # noqa: D102
        # remember the subparser to print usage in case no subverb is passed
        self._subparser = parser

        # get subverb extensions and let them add their arguments
        subverb_extensions = get_subverb_extensions()
        add_subparsers(
            parser, 'colcon clean', subverb_extensions,
            attribute='subverb_name')

    def main(self, *, context):  # noqa: D102
        # error: no subverb provided
        if context.args.subverb_name is None:
            print(self._subparser.format_usage())
            return 'Error: No subverb provided'
